import os
import sys
from os.path import join as opj
from operator import sub
from collections import OrderedDict
from amolkit import getEleInfo as gei 
from amolkit.stringmanip import updateAtomNames 
from amolkit import misc

def crdreadfmt(crdtype='extended'):
    if crdtype[0:3].lower() == 'ext':
        #fortran_format = '(2I10,2X,A8,2X,A8,3F20.10,2X,A8,2X,A8,F20.10)'
        indexdict={'srlno':[0,10],'resid':[10,20],'resn':[22,30],'aname':[32,40],
            'x':[40,60],'y':[60,80],'z':[80,100],'segn':[102,110],'segid':[112,120],'tfac':[120,140]}
    else:    
        #fortran_format = '(2I5,1X,A4,1X,A4,3F10.5,1X,A4,1X,A4,F10.5)'
        indexdict={'srlno':[0,5],'resid':[5,10],'resn':[11,15],'aname':[16,20],
            'x':[20,30],'y':[30,40],'z':[40,50],'segn':[51,55],'segid':[56,60],'tfac':[60,70]}
    return indexdict    
        
def pdbreadfmt(pdbtype='charmm'):
    indexdict={'srlno':[6,11],'aname':[12,16],'resn':[17,21],'chain':[22],'resid':[22,26],
                    'x':[30,38],'y':[38,46],'z':[46,54],'occu':[54,60],'tfac':[60,66],'segn':[72,76],'ele':[76,78],'charge':[78,80]}
    return indexdict

def crdwrtfmt(cd): 
    return "{:>10}{:>10}  {:<8}  {:<8}{:>20.10f}{:>20.10f}{:>20.10f}  {:<8}  {:<8}{:>20.10f}".format(cd["srlno"],
            cd["resid"],cd["resn"],cd["aname"],cd["coord"][0],cd["coord"][1],cd["coord"][2],cd["segn"],cd["resid"],0.0)

def pdbwrtfmt(pd):
    return "{:6s}{:5d} {:^4s}{:4s}  {:4d}    {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          {:4s}".format('ATOM',
            pd["srlno"],pd["aname"],pd["resn"],pd["resid"],pd["coord"][0],pd["coord"][1],pd["coord"][2],pd["occu"],pd["tfac"],pd["segn"])

def xyzwrtfmt(xd):
    return "%s %15.10f %15.10f %15.10f"%(xd["aname"],xd["coord"][0],xd["coord"][1],xd["coord"][2])

class Molecule():
    def __init__(self):
        self.atomindex_byId=OrderedDict()
        self.atomname_byIdx=OrderedDict()
        self.atomserial_byIdx=OrderedDict()
        self.atomresn_byIdx=OrderedDict()
        self.atomresn_byId=OrderedDict()
        self.atomresid_byIdx=OrderedDict()
        self.atomresid_byId=OrderedDict()
        self.atomcoord_byIdx=OrderedDict()
        self.atomcoord_byId=OrderedDict()
        self.atomchain_byIdx=OrderedDict()
        self.atomchain_byId=OrderedDict()
        self.atomoccu_byIdx=OrderedDict()
        self.atomoccu_byId=OrderedDict()
        self.atomtfac_byIdx=OrderedDict()
        self.atomtfac_byId=OrderedDict()
        self.atomsegn_byIdx=OrderedDict()
        self.atomsegn_byId=OrderedDict()
        self.atomele_byIdx=OrderedDict() 
        self.atomele_byId=OrderedDict() 
        self.atomcharge_byIdx=OrderedDict()
        self.atomcharge_byId=OrderedDict()
        self.atommass_byIdx=OrderedDict()
        self.atommass_byId=OrderedDict()
        self.bondindices=[]
        self.energy=None
        self.natoms=None
        self.nmoltoms=None
        self.totalcharge=None

    def readmol2(self,filename,istart=None,resid=None,bio=True):
        molblock = open(filename,"r").readlines()
        foundatm=False
        foundbnd=False
        if not istart: istart = 0
        ind=0
        for line in molblock:
            field = line.split()
            if len(field) !=0:
                if "@<TRIPOS>" in field[0].strip() and field[0].strip("@<TRIPOS>") not in ["MOLECULE","ATOM","BOND"]: break
            if len(field) != 0 and not foundatm:
                if field[0] == "@<TRIPOS>ATOM":
                    foundatm=True
                    foundbnd=False
                    continue
            elif len(field) != 0 and not foundbnd:
                if field[0] == "@<TRIPOS>BOND":
                    foundbnd = True
                    foundatm = False
                    continue
            if foundatm and len(field) >= 4:
                try:
                    ind += 1
                    srlno = int(istart) + ind
                    self.atomserial_byIdx[ind] = srlno
                    self.atomname_byIdx[ind] = field[1]
                    self.atomcoord_byIdx[ind] = list(map(lambda x:round(float(x),7),field[2:5]))
                    self.atomele_byIdx[ind] = field[5].split(".")[0]
                    try:
                        self.atomcharge_byIdx[ind]=float(field[-1])
                    except (ValueError,IndexError):
                        pass
                except (IndexError,ValueError):
                    # Reset the index back to previous value, because of error in atom line
                    ind = ind - 1
                    foundatm = False
                continue

            if foundbnd and len(field) == 4:
                self.bondindices.append([int(field[1]),int(field[2])])
                continue

        allnames = list(map(lambda x:x.upper(), self.atomname_byIdx.values())) 
        if len(allnames) ==  len(list(set(allnames))):
            self.atomindex_byId = {v: k for k, v in self.atomname_byIdx.items()}

        self.nmolatoms = len(self.atomname_byIdx)
        onlyatoms = [a for a in self.atomname_byIdx.values() if a[0:2].upper() != "LP" and a[0] != "D"]
        self.natoms = len(onlyatoms)
        self.totalcharge = int(round(sum(self.atomcharge_byIdx.values())))

    def readpdb(self,filename,istart=None,resid=None,bio=True):
        '''
        Parses the lines of a file object, if lines starts with ATOM or HETATM. 
        The names and the coordinates are extracted according to the PDB format.
        Providing user option for istart i.e. starting serial num and resid because these two may require adjustment.
        '''

        indexdict=pdbreadfmt('charmm')

        molblock = open(filename,"r").readlines()

        ind = 0
        for line in molblock:
            if line.startswith('ATOM') or line.startswith('HETATM'):
                ind += 1
                if istart == None:
                    srlno = int(line[indexdict['srlno'][0]:indexdict['srlno'][1]].strip())
                    self.atomserial_byIdx[ind] = srlno 
                else:
                    srlno = int(istart) + ind
                    self.atomserial_byIdx[ind] = srlno
                aname = line[indexdict['aname'][0]:indexdict['aname'][1]].strip()
                self.atomname_byIdx[ind] = aname 
                resn = line[indexdict['resn'][0]:indexdict['resn'][1]].strip()
                self.atomresn_byIdx[ind] = resn 
                chain = line[indexdict['chain'][0]].strip()
                self.atomchain_byIdx[ind] = chain
                if resid:
                    self.atomresid_byIdx[ind] = int(resid)
                else:
                    resid = int(line[indexdict['resid'][0]:indexdict['resid'][1]].strip())
                    self.atomresid_byIdx[ind] = resid
                x=round(float(line[indexdict['x'][0]:indexdict['x'][1]].strip()),7)
                y=round(float(line[indexdict['y'][0]:indexdict['y'][1]].strip()),7)
                z=round(float(line[indexdict['z'][0]:indexdict['z'][1]].strip()),7)
                self.atomcoord_byIdx[ind] = [x,y,z]
                occu = float(line[indexdict['occu'][0]:indexdict['occu'][1]].strip())
                self.atomoccu_byIdx[ind] = occu
                tfac = float(line[indexdict['tfac'][0]:indexdict['tfac'][1]].strip())
                self.atomtfac_byIdx[ind] = tfac
                segn = line[indexdict['segn'][0]:indexdict['segn'][1]].strip()
                self.atomsegn_byIdx[ind] = segn 
                ele = line[indexdict['ele'][0]:indexdict['ele'][1]].strip()
                if ele:
                    self.atomele_byIdx[ind] = ele
                else:
                    self.atomele_byIdx[ind] = gei.atomsymbol_by_atomname(aname,bio)
                charge = line[indexdict['charge'][0]:indexdict['charge'][1]].strip()
                if charge: 
                    self.atomcharge_byIdx[ind] = float(charge)

        # For single residue with all names different, populate the atomindex_byId
        allnames = list(map(lambda x:x.upper(), self.atomname_byIdx.values())) 
        if len(allnames) ==  len(list(set(allnames))):
            self.atomindex_byId = {v: k for k, v in self.atomname_byIdx.items()}

        self.nmolatoms=len(self.atomname_byIdx)
        onlyatoms = [a for a in self.atomname_byIdx.values() if a[0:2].upper() != "LP" and a[0] != "D"]
        self.natoms = len(onlyatoms)

    def readcrd(self,filename,istart=None,resid=None,bio=True):
        molblock = open(filename,"r").readlines()
        ind = 0
        crdtype = 'standard'
        for line in molblock:
            field = line.split()
            if field[0] == '*' or len(field) == 0:
                continue  
            if len(field) <= 2:
                self.nmolatoms = int(field[0])
                if field[-1].upper() == 'EXT': crdtype = 'extended'
                indexdict = crdreadfmt(crdtype)
                continue

            if len(field) >= 10 and field[0] != "*":
                ind += 1
                if istart == None:
                    srlno = int(field[0].strip())
                    self.atomserial_byIdx[ind] = srlno 
                else:
                    srlno = int(istart) + ind
                    self.atomserial_byIdx[ind] = srlno
                if resid:
                    self.atomresid_byIdx[ind] = int(resid)
                else:
                    resid = int(line[indexdict['resid'][0]:indexdict['resid'][1]].strip())
                    self.atomresid_byIdx[ind] = resid 
                resn = line[indexdict['resn'][0]:indexdict['resn'][1]].strip()
                self.atomresn_byIdx[ind] = resn 
                aname = line[indexdict['aname'][0]:indexdict['aname'][1]].strip()
                self.atomname_byIdx[ind] = aname 
                self.atomele_byIdx[ind] = gei.atomsymbol_by_atomname(aname,bio)
                x=round(float(line[indexdict['x'][0]:indexdict['x'][1]].strip()),7)
                y=round(float(line[indexdict['y'][0]:indexdict['y'][1]].strip()),7)
                z=round(float(line[indexdict['z'][0]:indexdict['z'][1]].strip()),7)
                self.atomcoord_byIdx[ind] = [x,y,z]
                segn = line[indexdict['segn'][0]:indexdict['segn'][1]].strip()
                self.atomsegn_byIdx[ind] = segn 
                tfac = float(line[indexdict['tfac'][0]:indexdict['tfac'][1]].strip())
                self.atomtfac_byIdx[ind] = tfac

        # sanity check
        if self.nmolatoms != ind:
            raise ValueError("Found %d coordinates in %r but should be %d coordinates." % (self.nmolatoms, filename, ind))

        allnames = list(map(lambda x:x.upper(), self.atomname_byIdx.values())) 
        if len(allnames) ==  len(list(set(allnames))):
            self.atomindex_byId = {v: k for k, v in self.atomname_byIdx.items()}

        onlyatoms = [a for a in self.atomname_byIdx.values() if a[0:2].upper() != "LP" and a[0] != "D"]
        self.natoms = len(onlyatoms)

    def readxyz(self,filename,istart=None,resid=None,bio=True):
        if not istart: istart = 0
        ind = 0
        molblock = open(filename,"r").readlines()
        for i,line in enumerate(molblock):
            field = line.split()
            if i == 0: self.nmolatoms = int(field[0]) 
            if i == 1: continue # Second line is comment
            if len(field) >= 4:
                ind += 1
                srlno = int(istart) + ind
                self.atomserial_byIdx[ind] = srlno
                self.atomname_byIdx[ind] = field[0]
                self.atomcoord_byIdx[ind] = list(map(lambda x:round(float(x),7), field[1:4]))
                self.atomele_byIdx[ind] = gei.atomsymbol_by_atomname(field[0],bio)

        # sanity check
        if self.nmolatoms != ind:
            raise ValueError("Found %d coordinates in %r but should be %d coordinates." % (self.nmolatoms, filename, ind))

        allnames = list(map(lambda x:x.upper(), self.atomname_byIdx.values())) 
        if len(allnames) ==  len(list(set(allnames))):
            self.atomindex_byId = {v: k for k, v in self.atomname_byIdx.items()}

        onlyatoms = [a for a in self.atomname_byIdx.values() if a[0:2].upper() != "LP" and a[0] != "D"]
        self.natoms = len(onlyatoms)
 
    @staticmethod
    def generateBonds(elementnames=[],coordinates=[]):
        """
        Expects list of names and coordinates of same length 
        """
        import numpy as np
        from amolkit import getEleInfo as gei 
    
        if elementnames and coordinates and len(elementnames) != len(coordinates):
            raise IndexError("names list and coordinate list don't match")
         
        coordinates = np.array(coordinates,dtype ='float32')
    
        delta = [None]*3
        for i in range(3):
           coord = coordinates[:,i]
           delta[i] = coord - coord.reshape((len(coord),1))
    
        dist = delta[0]**2 + delta[1]**2 + delta[2]**2 + 10. * np.identity(len(coordinates))
    
        radius = []
        for el in elementnames:
            radius.append(gei.covrad_by_atomname(el)) 
        radius = np.array(radius, dtype='float32')
    
        radcov = (radius + radius.reshape((len(radius),1)))*2
        bonded = np.array(np.where(dist - radcov<0.)).transpose()
        distmat = np.sqrt(dist[bonded[:,0],bonded[:,1]])
        
        # amolkit indices for molecules starts from 1
        bondindices=[]
        for b in bonded:
            bondindices.append(list(map(lambda x:x+1,b)))
    
        bondinfo={}
        bondinfo["radcov"]         = radcov
        bondinfo["pybondindices"]  = bonded
        bondinfo["distmat"]        = distmat
        bondinfo["bondindices"]    = bondindices
        return bondinfo 

    def genBonds(self):
        self.bondindices=self.generateBonds(list(self.atomele_byIdx.values()),list(self.atomcoord_byIdx.values()))["bondindices"]

    def autoupdateAtomNames(self):
        atomnames=updateAtomNames(list(self.atomname_byIdx.values()))
        for i in range(len(atomnames)):
            ind = i + 1
            self.atomname_byIdx[ind] = atomnames[i]

    def updateAtomNamesWith(self,atomnames):
        if not isinstance(atomnames,list) or len(atomnames) != len(list(self.atomname_byIdx.values())):
            raise IndexError ("Length of atomnames not same as molecule atoms.")
        for i in range(len(atomnames)):
            ind = i + 1
            self.atomname_byIdx[ind] = atomnames[i]

    def readMolecule(self,filename,istart=None,resid=None,bio=True):
        filestatus = misc.FileStatus(filename) 
        extension = misc.getExtension(filename)
        if not filestatus:
            raise RuntimeError ("Error reading file")
        if extension not in ["xyz","mol2","pdb","crd"]: 
            raise RuntimeError ("Cant read this file format")
        if extension == "pdb":  self.readpdb(filename,istart,resid,bio)
        if extension == "crd":  self.readcrd(filename,istart,resid,bio)
        if extension == "mol2": self.readmol2(filename,istart,resid,bio)
        if extension == "xyz":  self.readxyz(filename,istart,resid,bio)

def genMol(atomnames:list,atomcoords:list,**kwargs):
    if not isinstance(atomnames,list) and not isinstance(atomcoords,list):
        raise ValueError ("Provide list of atomnames and list of coords")
    if len(atomnames) != len(atomcoords):
        raise IndexError ("Length of atomnames and coords are not same.")
    
    resname = "RESN"; resid = 1; segname = "SEGN"
    if 'resname' in kwargs:
       resname = kwargs["resname"] 
    if 'resid' in kwargs:
       resid = kwargs["resid"]
    if 'segname' in kwargs:
       segname = kwargs["segname"]

    mol = Molecule()
    mol.nmolatoms = len(atomnames) 
    
    for i in range(len(atomnames)): 
        ind = i + 1
        mol.atomname_byIdx[ind] = atomnames[i]
        mol.atomcoord_byIdx[ind] = atomcoords[i] 
        mol.atomserial_byIdx[ind]= ind 
        mol.atomele_byIdx[ind]   = gei.atomsymbol_by_atomname(atomnames[i]) 
        mol.atomresn_byIdx[ind]  = resname
        mol.atomresid_byIdx[ind] = resid
        mol.atomsegn_byIdx[ind]  = segname 
    return mol

def coordline(molecule,ind,serialnum=None,resid=None):
    '''
    For index = ind in molinstance, this function assigns variables required 
    for writing coordinate line in crd/pdb/xyz/mol2
    Arguments:
        molinstance: instance of Molecule class with preloaded molecule information
        ind: Index for which information is required
        istart: In case the molecule in molinstance is not the first molecule, 
        then serial number of first atom can be provided.
        resid: User defined resid
    Returns:
        cpvar:: Dictionary containing all possible information extracted from molinstance[ind]
    '''

    cpvar={}
    if serialnum:
        cpvar["srlno"] = serialnum 
    else:    
        cpvar["srlno"] = molecule.atomserial_byIdx[ind] 

    cpvar["aname"] = molecule.atomname_byIdx[ind] 

    if molecule.atomresn_byIdx and ind in molecule.atomresn_byIdx.keys():
        cpvar["resn"] = molecule.atomresn_byIdx[ind] 
    else:    
        cpvar["resn"] = "RESN" 

    if molecule.atomchain_byIdx and ind in molecule.atomchain_byIdx.keys():
        cpvar["chain"] = molecule.atomchain_byIdx[ind] 
    else:    
        cpvar["chain"] = "" 

    if resid:
        cpvar["resid"] = resid 
    elif molecule.atomresid_byIdx and ind in molecule.atomresid_byIdx.keys():
        cpvar["resid"] = molecule.atomresid_byIdx[ind] 
    else:    
        cpvar["resid"] = 1 

    cpvar["coord"] = molecule.atomcoord_byIdx[ind] 

    if molecule.atomoccu_byIdx and ind in molecule.atomoccu_byIdx.keys():
        cpvar["occu"] = molecule.atomoccu_byIdx[ind] 
    else:    
        cpvar["occu"] = 0.0  

    if molecule.atomtfac_byIdx and ind in molecule.atomtfac_byIdx.keys():
        cpvar["tfac"] = molecule.atomtfac_byIdx[ind] 
    else:    
        cpvar["tfac"] = 0.0  

    if molecule.atomsegn_byIdx and ind in molecule.atomsegn_byIdx.keys():
        cpvar["segn"] = molecule.atomsegn_byIdx[ind] 
    else:    
        cpvar["segn"] = "SEGN" 

    if molecule.atomele_byIdx and ind in molecule.atomele_byIdx.keys():
        cpvar["ele"] = molecule.atomele_byIdx[ind] 
    else:    
        cpvar["ele"] = "" 

    if molecule.atomcharge_byIdx and ind in molecule.atomcharge_byIdx.keys():
        cpvar["charge"] = molecule.atomcharge_byIdx[ind] 
    else:    
        cpvar["charge"] = 0.0 
    return cpvar 
 
def writecrd(molecule,fileout=None,nolpd=False,comment=None):
    ocrdfile = open(fileout,"w")
    if comment:
        ocrdfile.write("* %s\n"%(comment))
    else:
        ocrdfile.write("* Generated by amolkit\n")
    ocrdfile.write("* \n")
    if not nolpd:
        ocrdfile.write("%10i  EXT\n" %(molecule.nmolatoms))
        for i in range(molecule.nmolatoms):
            ind=i+1
            cdict = coordline(molecule,ind)
            ocrdfile.write(crdwrtfmt(cdict)+"\n")
        ocrdfile.close()
    else:
        serialnum = 0
        ocrdfile.write("%10i  EXT\n" %(molecule.natoms))
        for i in range(molecule.nmolatoms):
            ind=i+1
            checkatm = molecule.atomname_byIdx[ind]
            if checkatm[0:1] == "D" or checkatm[0:2] == "LP":
                serialnum = serialnum
            else:    
                serialnum = serialnum + 1
            cdict = coordline(molecule,ind,serialnum=serialnum)
            ocrdfile.write(crdwrtfmt(cdict)+"\n")
        ocrdfile.close()
    prntscr = " ".join((str(fileout),"created"))
    print(prntscr) 

def writepdb(molecule,fileout=None,nolpd=False,comment=None):
    #molecule = molecule()
    opdbfile = open(fileout,"w")
    if comment:
        opdbfile.write("REMARK   %s\n"%(comment))
    else:
        opdbfile.write("REMARK   1 Generated by amolkit\n")
    if not nolpd:
        for i in range(molecule.nmolatoms):
            ind=i+1
            cdict = coordline(molecule,ind)
            opdbfile.write(pdbwrtfmt(cdict)+"\n")
    else:
        serialnum = 0
        for i in range(molecule.nmolatoms):
            ind=i+1
            checkatm = molecule.atomname_byIdx[ind]
            if checkatm[0:1] == "D" or checkatm[0:2] == "LP":
                serialnum = serialnum
            else:    
                serialnum = serialnum + 1
            pdict = coordline(molecule,ind,serialnum=serialnum)
            opdbfile.write(pdbwrtfmt(pdict)+"\n")
    opdbfile.write("TER\n")
    opdbfile.write("END\n")
    opdbfile.close()
    prntscr = " ".join((str(fileout),"created"))
    print(prntscr) 

def writexyz(molecule,fileout=None,nolpd=False,comment=None):
    #molecule = molecule()
    if fileout == None:
       fileout = "molecule.xyz"
       print ("No output xyzfile name provided. Writing in molecule.xyz") 

    oxyzfile = open(fileout,"w")
    if not nolpd:
        oxyzfile.write("%10i \n" %(molecule.nmolatoms))
        if comment:
            oxyzfile.write("%s\n"%(comment))
        else:
            oxyzfile.write("Generated by amolkit\n")
        for i in range(molecule.nmolatoms):
            ind=i+1
            xdict = coordline(molecule,ind)
            oxyzfile.write(xyzwrtfmt(xdict)+"\n")
        oxyzfile.close()
    else:
        serialnum = 0
        oxyzfile.write("%10i\n" %(molecule.natoms))
        if comment:
            oxyzfile.write("%s\n"%(comment))
        else:
            oxyzfile.write("Generated by amolkit\n")
        for i in range(molecule.nmolatoms):
            ind=i+1
            checkatm = molecule.atomname_byIdx[ind]
            if checkatm[0:1] == "D" or checkatm[0:2] == "LP":
                serialnum = serialnum
            else:    
                serialnum = serialnum + 1
            xdict = coordline(molecule,ind,serialnum=serialnum)
            oxyzfile.write(xyzwrtfmt(xdict)+"\n")
        oxyzfile.close()
    prntscr = " ".join((str(fileout),"created"))
    print(prntscr) 

class GenMoleculeRDkit():
    import numpy as np
    try:
        from rdkit import Chem
        from rdkit.Chem import AllChem
    except ImportError:
        logging.warning('RDKit not found!')


    def __init__(self,topology,resname,resitopfile):
        """
        """
        self.topology = topology()
        self.topology.loadCharmmTopology(resname,resitopfile)

    def genRDMolTop(self):
        self.mol=Chem.RWMol()
        atid={}
        idat={}
        atomicnumber_by_atomname
        for i, aname in topology.atomname_byIdx.items():
            if aname[0]=="D" or aname[0:2] =="LP": continue
            atomicnumber = gei.atomicnumber_by_atomname(aname) 
            atomsymbol = gei.atomsymbol_by_atomname(aname) 
            idx = m.AddAtom(Chem.Atom(atomicnumber))
            atid.update({atomsymbol:idx})
            idat.update({idx:atomsymbol})
        for bond in self.bonds:
            self.mol.AddBond(atid[bond[0]], atid[bond[1]],Chem.BondType.SINGLE)
        try:
            Chem.SanitizeMol(self.mol)
        except ValueError:
            pass
    
        #self.rdmolnrs=idat

    def gen3dRDkit(self):
        AllChem.EmbedMolecule(mol,useExpTorsionAnglePrefs=True,useBasicKnowledge=True)
        AllChem.MMFFOptimizeMolecule(mol)




def getPenaltyfromString(comment,penaltyId="PENALTY"):
    #Penalty is in comments. Its value is usually provided after word "penalty"
    #Will be sucessful if comment is penalty x.xx or penalty= x.xx or penalty = x.xx
    penalty = 0.0 
    comment = ""
    cmtuplist = comment.upper().split()
    cmtuplist = list(map(lambda x:x[0:7],cmtuplist))
    cmtlist = comment.split()
    penaltyId = penaltyId.upper()

    if penaltyId in cmtuplist:
        penpos = cmtuplist.index(penaltyId)
        findpenalty = cmtlist[penpos:] 
        try:
            penalty = round(float(findpenalty[1]),4)
            comment = " ".join(cmtlist[0:penpos]+cmtlist[penpos+2:])
        except (ValueError,IndexError):
            try:
                penalty = round(float(findpenalty[2]),4)
                comment = " ".join(cmtlist[0:penpos]+cmtlist[penpos+3:])
            except (ValueError,IndexError):
                penalty = 0.0
                comment = ""
    return penalty,comment

def removeSym(name):
    name = re.sub('[^A-Za-z0-9]+', '', name)
    name = name[0:6].strip()
    if len(name) < 4: name=name+"".join(["F"]*(4-len(name))) 
    return (name)

def stripname(atomname):
    """
    This function extracts first set of continous alphabetic characters from atomname.
    Replaces all special characters by 0. Skips any digit before alphabetic character.
    Trims the atomname till next occurance of digit, if any.
    Warning: If atomname consists of only symbols and numbers, function will return blank
    """

    import re

    tmpatomname=re.sub('[^A-Za-z0-9]+', '0', atomname)
    charfound=False

    for i,x in enumerate(tmpatomname):
        if x.isdigit() and not charfound:
            start = i
            end = i 
            continue
        elif not x.isdigit() and not charfound:
            start = i
            end = i + 1
            charfound=True
        elif not x.isdigit() and charfound:
            end = i + 1
        elif x.isdigit() and charfound:
            end = i 
            break

    atomname=tmpatomname[start:end]
    return (atomname)

def renumber(prevname,bio=True):
    #atomsymbol=atomsymbol_by_atomname(atomname,bio)
    #atomname=list(map(lambda x:gei.atomsymbol_by_atomname(x[0:4].strip(),bio),prevname))
    atomname=list(map(lambda x:x[0:4].strip(),prevname))
    collect={}
    newname=[None]*len(atomname)
    for i,a in enumerate(atomname):
        if a not in collect.keys():
            collect[a]=[i]
        else:
            collect[a].append(i) 
    for key,value in collect.items():
        i=0
        if len(value)==1: 
            newname[value[0]]=key
            continue
        if len(str(len(value)))>=5:
            raise Exception("Cannot handle molecules with more than 9999 non-unique atom names")
        if len(key)+len(str(len(value)))>5:
            trunc=5-len(str(len(value)))
            #print (trunc)
            newkey=key[0:trunc]
            if newkey in collect.keys():
                i=len(collect[newkey])
        else:
            newkey=key
        for v in value:
            i=i+1
            newname[v]=newkey+str(i)
                   
    return newname       

def updateAtomNames(iatomnames):
    update_names=renumber(list(iatomnames.values()))
    for i,nam in enumerate(update_names):
        iatomnames[i+1]=nam
    return iatomnames

def split_line(line, max_length=60):
    """Splits a line into multiple lines if its length exceeds max_length, preserving continuous word."""

    if len(line) <= max_length:
        return [line]  

    splitline = []
    words = line.split()
    current_line = ""

    for word in words:
        if len(current_line) + len(word) + 1 > max_length:  
            splitline.append(current_line.strip())
            current_line = word
        else:
            current_line += " " + word

    splitline.append(current_line.strip())

    return splitline



from amolkit import getEleInfo as gei 
#class PsfAtom:
#    def __init__(self, atom_name, atom_index, atom_type, atom_charge, atom_mass, atom_resname, atom_resid, atom_segn, atom_qlp):
#        self.name = atom_name
#        self.index = atom_index
#        self.type = atom_type
#        self.charge = atom_charge
#        self.mass = atom_mass
#        self.resname = atom_resname
#        self.resid = atom_resid
#        self.segn = atom_segn
#        self.qlp = atom_qlp
#        self.alpha = None
#        self.thole = None
#
#class Psf:
#    def __init__(self):
#        self.atoms = {}
#        self.bond_indices = []
#        self.lpbond_indices = []
#        self.drudebond_indices = []
#        self.angle_indices = []
#        self.dihedral_indices = []
#        self.improper_indices = []
#        self.donor_indices = []
#        self.acceptor_indices = []
#        self.cmap_indices = []
#        self.lpics = []
#        self.anisotropies = []
#        self.drudebonds = []
#        self.groups = []
#
#    def add_atom(self, atom_name, atom_index, atom_type, atom_charge, atom_mass, atom_resname, atom_resid, atom_segn, atom_qlp, atom_alpha=None, atom_thole=None):
#        atom = PsfAtom(atom_name, atom_index, atom_type, atom_charge, atom_mass, atom_resname, atom_resid, atom_segn, atom_qlp)
#        atom.alpha = atom_alpha
#        atom.thole = atom_thole
#        self.atoms[atom_index] = atom
#
#    def add_bond(self, atom_a, atom_b):
#        self.bond_indices.append([atom_a, atom_b])

class Atom():
    def __init__(self,index,name=None,serial=None,atype=None,dtype=None,symbol=None,mass=None,charge=None,
                 Z=None,x=None,y=None,z=None,resn=None,resid=None,segn=None,chain=None,
                 alpha=None,thole=None,occu=None,tfac=None,penalty=None,comment=None):
        self.index   = index
        self.name    = name 
        self.serial  = serial 
        self.atype   = atype 
        self.dtype   = dtype 
        self.symbol  = symbol 
        self.Z       = Z 
        self.mass    = mass 
        self.charge  = charge 
        self.alpha   = alpha
        self.thole   = thole
        self.resn    = resn 
        self.resid   = resid 
        self.chain   = chain 
        self.segn    = segn 
        self.x       = x 
        self.y       = y 
        self.z       = z 
        self.occu    = occu 
        self.tfac    = tfac 
        self.penalty = penalty
        self.comment = comment

    def GetName(self):
        return self.name

    def GetSymbol(self,bio=True):
        self.symbol = gei.atomsymbol_by_atomname(self.name,bio)
        return self.symbol

    def GetSymbolbyAtomicNumber(self): 
        self.symbol = gei.atomsymbol_by_atomicnumber(self.Z) #atomicnumber) 
        return self.symbol

    def GetAtomicNumber(self,bio=True):
        self.Z = gei.atomicnumber_by_atomname(self.name,bio)
        return self.Z

    def GetAtomicMass(self,bio=True):
        self.mass = gei.atomicmass_by_atomname(self.name,bio)
        return self.mass

    def GetCovRad(self,bio=True):
        self.radius = gei.covrad_by_atomname(self.name,bio)
        return self.radius

    def GetCoordinate(self):
        self.coord = [self.x,self.y,self.z]
        return self.coord

class ElementProp():
   def __init__(self, Z, radius, max_bonds, r, g, b):
      self.Z = Z
      self.radius = radius
      self.max_bonds = max_bonds
      self.color = np.array([r,g,b], dtype='float32')
   def __str__(self):
      format_string = ' '.join(['{:4d}','{:9f}','{:4d}','{:9f}'*3])
      return format_string.format(self.Z, self.radius, self.max_bonds, *self.color)

class AssignElementProp(ElementProp):
   def __init__(self): 
      self.PSE = {}
      self.ATSYM = {}
      for atsym in ei.AtomOrder:
         self.PSE[atsym.upper()] = ElementProp(ei.AtomicNumber[atsym],ei.CovalentRadius[atsym],ei.BondOrder[atsym],*ei.AtomColor[atsym])
         self.ATSYM[ei.AtomicNumber[atsym]] = atsym



import logging
try:
    from rdkit import Chem
    from rdkit.Chem import AllChem
except ImportError:
    logging.warning('RDKit not found!')
import numpy as np
from collections import OrderedDict

class MolTransform():
    __transformtypes={2:'bond',3:'angle',4:'dihedral'}
    def __init__(self,m,resname,rescharge,coor,atomnames,atomnrmap,bondnames,changetype,atomlist,targetvalue=None,targetrange=None,returns=None,pathfilebase=None,writepdb=False,writepy=False,multiplicity=1):
        self.m=m
        self.resname=resname
        self.rescharge=rescharge
        self.multiplicity=multiplicity
        self.atomnames=atomnames
        self.atomnrmap= atomnrmap
        self.bonds=bondnames
        self.coor=coor
        self.changetype=changetype
        self.scanatomnames=atomlist
        self.scanatomnr=[self.atomnrmap[i][0] for i in self.scanatomnames]
        self.targetvalue=targetvalue
        self.targetrange=targetrange
        self.returns=returns
        self.pfbase=pathfilebase
        self.writepy=writepy
        self.writepdb=writepdb

        if not self.__ctrl_changetype():
            return(None)
        if not self.__ctrl_bonds():
            return(None)
        if not self.__ctrl_length():
            return(None)
        if not self.__ctrl_coor():
            return(None)
        if not self.targetvalue:
            if not self.__ctrl_range():
                return(None)
        elif self.targetvalue:
            if type(self.targetvalue) not in [str,int,float]:
                logging.error('Target value keyword excepts a single value')
                return(None)
            else:
                try:
                    self.targetvalue=float(self.targetvalue)
                except ValueError:
                    logging.error("Cannot convert target value to float")
                    return(None)
        else:
            logging.error('Provide either target value or target range not both.')
            return(None)

        if not self.returns:
            if not self.pfbase:
                logging.error('Either provide a base filename (prepended with the path) or request to return the generated coordinate objects.')
        self.__transform()

    def __ctrl_changetype(self):
        #Sanity check
        typefromlist=self.__transformtypes[len(self.scanatomnr)]
        if typefromlist != self.changetype:
            logging.error('The transformation type %s requies % atoms.' %(self.changetype,len(self.scanatomnr)))
            return(None)
        if self.changetype not in self.__transformtypes.values():
            logging.error('Transformation type %s is not supported, Please choose from:%'%(self.changetype,','.join(self.__transformtypes.values())))
            return(None)
        return(True)

    def __ctrl_bonds(self):
        natom={'bond':2,'angle':3,'dihedral':4}
        bonded_groups=[self.scanatomnames[0:2]]
        if self.changetype == 'angle':
            bonded_groups+=[self.scanatomnames[1:3]]
        elif self.changetype == 'dihedral':
            bonded_groups+=[self.scanatomnames[1:3]]
            bonded_groups+=[self.scanatomnames[2:]]
        for bonded in bonded_groups:
            if len(bonded) != 2:
                logging.error("The atomlist should contain %s atoms for transformation type %s" %(natom[self.changetype],self.changetype))
                return False
            if bonded not in self.bonds:
                if [bonded[1],bonded[0]] not in self.bonds:
                    logging.error("%s %s atoms are not bonded. Cannot carry on with the transformation."%(bonded[0],bonded[1]))
                    return False
        return True

    def __ctrl_length(self):
        if len(self.atomnames) != len(self.coor):
            logging.error('Topology has %s atoms, but %s coordinates is passed.' %(len(self.atomnames),len(self.coor)))
            return False
        return True

    def __ctrl_coor(self):
        if not isinstance(self.coor,np.ndarray):
            self.coor=np.array(self.coor)
        nrs,names=self.coor[:,0:1],self.coor[:,1:2]
        try:
            [float(i) for i in names]
        except ValueError:
            return True
        else:
            logging.error('Please pass the coordinates in this format: [[Nr, AtName,x,y,z],[],[]]')
            return False
        return True

    def __ctrl_range(self):
        if self.targetrange is None:
            logging.error('Either provide a single target value or provide a range using the keyword arugments targetvalue or targetrange.')
            return False
        if len(self.targetrange) != 3:
            logging.error('The target range must have 3 elements, start, end , increment')
            return False
        try:
            self.targetrange=[float(i) for i in self.targetrange]
        except ValueError:
            logging.error('The target range contains characters. Please pass only numbers.')
            return False
        return True

    def __transformcore(self,m,coor,changetype,atomlist,value):
        #RDKit mols are C++ objects, pointers to that object created. So we either do a deepy copy or create another molecule to work with temporarily, otherwise the original molecule is modified.
        nrname=coor[:,0:2]
        tmp=Chem.Mol(m)
        tmp.RemoveAllConformers()
        AllChem.EmbedMultipleConfs(tmp,1)
        funcs={'bond':Chem.rdMolTransforms.SetBondLength,'angle':Chem.rdMolTransforms.SetAngleDeg,'dihedral':Chem.rdMolTransforms.SetDihedralDeg}
        #RDKit numbers start from 0
        atomlist=[int(i)-1 for i in atomlist]
        value=float(value)
        conf=tmp.GetConformer()
        for i in range(tmp.GetNumAtoms()):
            conf.SetAtomPosition(i,[float(coor[i][2]),float(coor[i][3]),float(coor[i][4])])
        funcs[changetype](conf,*atomlist,value)
        a=conf.GetPositions()
        a=np.hstack((nrname,a))
        return(a)

    def __transform(self):
        if self.targetvalue:
            values=[self.targetvalue]
        elif self.targetrange:
            values=mc.frange(self.targetrange,rdigit=3,step=True)
        returncoor={}
        if self.writepy:
            pyoutn=self.pfbase+'_xyz.py'
            pyout=open(pyoutn,'w')
            objname='scan_'
            pyout.write('basename="%s" \n' %objname)
            pyout.write('scanrange=[')
            pyvalues=[]
            for i in values:
                i='_'.join(str(i).split('.'))
                if '-' in i:
                    i='n'.join(str(i).split('-'))
                pyvalues.append(i)
                pyout.write('"%s",' %i)
            pyout.write(']\n')
        elif self.writepdb:
            pdbns=[]
        for val in range(len(values)):
            mol=Chem.Mol(self.m)
            newcoor=self.__transformcore(mol,self.coor,self.changetype,self.scanatomnr,values[val])
            if self.returns:
                returncoor.update({pyvalues[val]:newcoor})
            else:
                if self.writepdb:
                    pdbn=self.pfbase+'_'+str(values[val])+'.pdb'
                    pdbns.append(pdbn)
                    with open(pdbn,'w') as pdbout:
                        for coor in newcoor:
                            pdbout.write(cw.crdline(coor,returntype='pdb'))
                if self.writepy:
                    pyout.write("%s%s='''\n"%(objname,pyvalues[val]))
                    pyout.write("%d %d \n" %(self.rescharge,self.multiplicity))
                    #Only underscore within objectnames
                    for coor in newcoor:
                        #The crdline returns with a new line
                        pyout.write(cw.crdline(coor,returntype='xyz'))
                    pyout.write('\nnoreorient\nnocom\n')
                    pyout.write("'''\n")
                    pyout.write("\n")
        if self.returns:
            return(returncoor)
        if self.writepy:
            pyout.close()
            return(pyoutn)
        if self.writepdb:
            return(pdbns)

def align(positions,center=False):
    positions = positions - positions.min(axis=0)
    if center:
        origin = np.array([0.,0.,0.],dtype='float32')
        origin = sum(positions)/len(positions)
        positions = positions - origin
    return positions



import os,sys

def getExtension(filein):
    extension = filein.split('.')[-1].lower()
    return extension

def FileStatus(filein): 
    status = True
    if not os.path.exists(filein): status=False
    if os.stat(filein).st_size==0: status=False
    if os.stat(filein).st_size/(1024 * 1024)>50: status=False
    return status



import sys
import math
import argparse
from collections import OrderedDict
from amolkit import genic
from amolkit import getEleInfo as gei 
from amolkit.stringmanip import getPenaltyfromString
from amolkit.stringmanip import split_line 

class Topology():
    def __init__(self): 
        """
        Constructor for the Topology class.
        """
    
    @staticmethod
    def readCharmmTopology(resname,resitopfile,bio=True):
        '''
        Load the features of the residue based on topology file.
    
        Note::
        field[1] is atomname
        field[2] is atomtype
        if bio is True then utilize gei.atomicmass_by_atomname upon atomtype
        to get atomicmass or symbol. This way I am not limited to predefined atomtypes in EleInfo.
    
        atommass_byId[field[1]] = gei.atomicmass_by_atomname(field[2],bio)
        atomele_byId[field[1]] = gei.atomsymbol_by_atomname(field[2],bio)
    
        But if it false, then I should ideally get atomicmass by atomtype
        and in case an atomtype is missing, I should allow user to supply new atomtype and its correspondence.
    
        # Give some thought if bio is False. In that case, atomtype = MAG will not be converted to Mg
        # Thus it would not find Mg
        '''
    
        topology = {}
        topology["resname"]            = resname
        topology["fftype"]             = 'additive' 
        topology["rescharge"]          = None 
        topology['ntopatoms']          = None
        topology['natoms']             = None
        topology['first_byId']         = None
        topology['last_byId']          = None
        topology['atomindex_byId']     = OrderedDict() 
        topology['atomtype_byId']      = OrderedDict()
        topology['atomcharge_byId']    = OrderedDict()
        topology['atompenalty_byId']   = OrderedDict()
        topology['atommass_byId']      = OrderedDict()
        topology['atomele_byId']       = OrderedDict()
        topology['atomalpha_byId']     = OrderedDict()
        topology['atomthole_byId']     = OrderedDict()
        topology['atomdrudetype_byId'] = OrderedDict()
        topology['atomname_byIdx']     = OrderedDict()
        topology['atomtype_byIdx']     = OrderedDict()
        topology['atomcharge_byIdx']   = OrderedDict()
        topology['atompenalty_byIdx']  = OrderedDict()
        topology['atommass_byIdx']     = OrderedDict()
        topology['atomele_byIdx']      = OrderedDict()
        topology['atomalpha_byIdx']    = OrderedDict()
        topology['atomthole_byIdx']    = OrderedDict()
        topology['atomdrudetype_byIdx']= OrderedDict()
        topology['atomsinring']        = OrderedDict()
        topology['groups']             = []  
        topology['atomnames']          = []    
        topology['bonds']              = []
        topology['lpbonds']            = []
        topology['angles']             = []
        topology['dihedrals']          = []
        topology['impropers']          = []
        topology['lpics']              = []
        topology['anisotropies']       = []
        topology['donors']             = []
        topology['acceptors']          = []
        topology['cmaps']              = []
        topology['atomcharges']        = []
    
        try:
            filein=open(resitopfile, "r")
        except IOError:
            raise IOError('Could not access the ' + resitopfile + ' topology file')
    
        found=False
        ring = 0
        nat = 0
    
        for nl,line in enumerate(filein):
            #Strip the line, remove the comment lines, and split using whitespace
            field = line.strip().split("!")[0].split()
            
            fieldup = list(map(lambda x:x.upper(),field))
    
            comment = "!".join(line.strip().split("!")[1:]) 
    
            if len(fieldup) >= 1:
                ftype = fieldup[0][0:4]
                if found and ftype == "END": break
                if ftype in ["RESI","PRES"] and fieldup[1] !=  topology['resname'].upper():
                    if found:
                        break
                    else:
                       found = False
                elif ftype in ["RESI","PRES"] and fieldup[1] ==  topology['resname'].upper():
                    found = True
                    try:
                        topology["rescharge"] = round(float(field[2]),3)
                    except (ValueError,IndexError):
                        raise ValueError('Beware:: Error in detecting residue charge of '+topology['resname']+'.')
    
            if found and len(fieldup) > 0:
                ftype = fieldup[0][0:4]
                if ftype not in ["RING","GROU","ATOM","BOND","DOUB","TRIP","ANGL","DIHE","IMPR","DONO","ACCE","CMAP","LONE","ANIS"]: continue
                if ftype == "RING":
                    ring=ring+1
                    #topology['atomsinring'][ring]=field[1:]
                    topology['atomsinring'][ring] = fieldup[1:]
                    continue
                elif ftype == "GROU":
                    topology['groups'].append([nat,0,0])
                    continue
                elif ftype == "ATOM" and len(fieldup) > 3: 
                    nat = nat + 1
                    topology['atomindex_byId'][fieldup[1]] = nat
                    topology['atomname_byIdx'][nat] = fieldup[1]    # Atomname stored as in str
                    topology['atomtype_byId'][fieldup[1]] = fieldup[2]  # Atomtype capitalized before storing
                    topology['atomtype_byIdx'][nat] = fieldup[2]  # Atomtype capitalized before storing
                    topology['atomcharge_byId'][fieldup[1]] = float(field[3])
                    topology['atomcharge_byIdx'][nat] = float(field[3])
                    # Using atomicmass_by_atomname, but supplying atomtype
                    # Because atomicmass_by_atomtype does a fixed conversion
                    topology['atommass_byId'][fieldup[1]] = gei.atomicmass_by_atomname(field[2],bio)
                    topology['atommass_byIdx'][nat] = gei.atomicmass_by_atomname(field[2],bio)
                    topology['atomele_byId'][fieldup[1]] = gei.atomsymbol_by_atomname(field[2],bio)
                    topology['atomele_byIdx'][nat] = gei.atomsymbol_by_atomname(field[2],bio)
                    topology['atomcharges'].append(float(field[3]))
                    penalty = comment.split()[0] if comment else str(0.0)
                    topology['atompenalty_byId'][fieldup[1]] = penalty 
                    topology['atompenalty_byIdx'][nat] = penalty 
    
                    if topology['atomnames'] and fieldup[1] in topology['atomnames']:
                        raise ValueError("Topology contains repeated atomname %s. Unable to load."%(field[1]))
                    else:
                        topology['atomnames'].append(fieldup[1]) 
                    
                    if any(list(map(lambda x:x in fieldup, ["ALPHA","THOLE","TYPE"]))): 
                        topology['atomalpha_byId'][fieldup[1]]=None
                        topology['atomalpha_byIdx'][nat]=None
                        topology['atomthole_byId'][fieldup[1]]=None
                        topology['atomthole_byIdx'][nat]=None
                        topology['atomdrudetype_byId'][fieldup[1]]='DRUD'
                        topology['atomdrudetype_byIdx'][nat]='DRUD'
                        topology["fftype"] = "drude"
    
                    if "ALPHA" in fieldup:
                        topology['atomalpha_byId'][fieldup[1]]=float(field[fieldup.index("ALPHA")+1])
                        topology['atomalpha_byIdx'][nat]=float(field[fieldup.index("ALPHA")+1])
                    if "THOLE" in fieldup:
                        topology['atomthole_byId'][fieldup[1]]=float(field[fieldup.index("THOLE")+1])
                        topology['atomthole_byIdx'][nat]=float(field[fieldup.index("THOLE")+1])
                    if "TYPE" in fieldup:
                        topology['atomdrudetype_byId'][fieldup[1]]=fieldup[fieldup.index("TYPE")+1]
                        topology['atomdrudetype_byIdx'][nat]=fieldup[fieldup.index("TYPE")+1]
    
                    continue
    
                #Bond list dictionary of each atom is populated.
                elif ftype in ["BOND","DOUB","TRIP"] and (len(fieldup)-1) % 2 == 0:
                    for i in range(1,len(fieldup),2):
                        ba = fieldup[i].strip("+-")
                        bb = fieldup[i+1].strip("+-")
                        topology['bonds'].append([ba,bb])
    
                        if ba.startswith('+'): topology['first_byId'] = ba
                        if bb.startswith('+'): topology['first_byId'] = bb
                        if ba.startswith('-'): topology['last_byId'] = ba
                        if bb.startswith('-'): topology['last_byId'] = bb
    
                        if 'LP' in [ba[0:2], bb[0:2]]:
                            topology['lpbonds'].append([ba,bb])
    
                elif ftype in ["BOND","DOUB","TRIP"] and (len(fieldup)-1) % 2 != 0:
                     raise ValueError('Topology file BOND line has the wrong format. It should be multiples of two! BOND At1 At2 At2 At3 At1 At3 etc.')
    
                #Angle list dictionary.
                elif ftype == "ANGL" and (len(fieldup)-1) % 3 == 0:
                    for i in range(1,len(fieldup),3):
                        aa = fieldup[i].strip("+-")
                        ab = fieldup[i+1].strip("+-")
                        ac = fieldup[i+2].strip("+-")
                        topology['angles'].append([aa,ab,ac])
                    continue
                elif ftype == "ANGL" and (len(fieldup)-1) % 3 != 0:
                    raise ValueError('Topology file ANGLE line has the wrong format. Ignored reading.')
                    continue
    
                #Dihedral list dictionary.
                elif ftype == "DIHE" and (len(fieldup)-1) % 4 == 0:
                    for i in range(1,len(fieldup),4):
                        da = fieldup[i].strip("+-")
                        db = fieldup[i+1].strip("+-")
                        dc = fieldup[i+2].strip("+-")
                        dd = fieldup[i+3].strip("+-")
                        topology['dihedrals'].append([da,db,dc,dd])
    
                elif ftype == "DIHE" and (len(fieldup)-1) % 4 != 0:
                    raise ValueError('Topology file DIHEDRAL line has the wrong format. Ignored reading.')
    
                #Improper list dictionary.
                elif ftype == "IMPR" and (len(fieldup)-1) % 4 == 0:
                    for i in range(1,len(fieldup),4):
                        ima = fieldup[i].strip("+-")
                        imb = fieldup[i+1].strip("+-")
                        imc = fieldup[i+2].strip("+-")
                        imd = fieldup[i+3].strip("+-")
                        topology['impropers'].append([ima,imb,imc,imd])
                elif ftype == "IMPR" and (len(fieldup)-1) % 4 != 0:
                    raise ValueError('Topology file IMPROPER line has the wrong format. Ignored reading.')
    
                #Donor list dictionary.
                elif ftype == "DONO" and (len(fieldup)-1) % 2 == 0:
                    for i in range(1,len(fieldup),2):
                        topology['donors'].append([fieldup[i+1],fieldup[i]])
    
                elif ftype == "DONO" and (len(fieldup)-1) % 2 != 0:
                    raise ValueError('Topology file DONOR line has the wrong format. Ignored reading.')
    
                #Acceptor list dictionary.
                elif ftype == "ACCE" and (len(fieldup)-1) % 2 == 0:
                    for i in range(1,len(fieldup),2):
                        topology['acceptors'].append([fieldup[i],fieldup[i+1]])
    
                elif ftype == "ACCE" and (len(fieldup)-1) == 1:
                    topology['acceptors'].append([fieldup[-1],""])
    
                elif ftype == "ACCE" and (len(fieldup)-1) != 1 or ftype == "ACCE" and (len(fieldup)-1) % 2 != 0:
                    raise ValueError('Topology file ACCEPTOR line has the wrong format. Ignored reading.')
    
                elif ftype == "CMAP" and (len(fieldup)-1) % 8 == 0:
                    for i in range(1,len(fieldup),8):
                        ca = fieldup[i].strip("+-")
                        cb = fieldup[i+1].strip("+-")
                        cc = fieldup[i+2].strip("+-")
                        cd = fieldup[i+3].strip("+-")
                        ce = fieldup[i+4].strip("+-")
                        cf = fieldup[i+5].strip("+-")
                        cg = fieldup[i+6].strip("+-")
                        ch = fieldup[i+7].strip("+-")
                        topology['cmaps'].append([ca,cb,cc,cd,ce,cf,cg,ch])
    
                elif ftype == "CMAP" and (len(fieldup)-1) % 8 != 0:
                    raise ValueError('Topology file CMAP line has the wrong format. Ignored reading.')
                
                # topology['atomindex_byId'] will change later according to inclusion of 
                # Lone pair list dictionary
                elif ftype == "LONE": 
                    if fieldup[1][0:4] in ["RELA","BISE"]:
                        try:
                            hlist = fieldup[2:6]
                            vlist = []
                            for ilp in range(len(fieldup)): 
                                if fieldup[ilp][0:4] == "DIST" and fieldup[1][0:4] == "RELA":
                                    vlist.append(float(fieldup[ilp+1]))
                                if fieldup[ilp][0:4] == "DIST" and fieldup[1][0:4] == "BISE":
                                    vlist.append(-(float(fieldup[ilp+1])))
                                if fieldup[ilp][0:4] == "ANGL":
                                    vlist.append(float(fieldup[ilp+1]))
                                if fieldup[ilp][0:4] == "DIHE":
                                    vlist.append(float(fieldup[ilp+1]))
                            topology['lpics'].append([fieldup[1][0:4],hlist,vlist])
                        except (IndexError, ValueError,KeyError):    
                            raise ValueError('Topology file LONEPAIR line has the wrong format. Ignored reading.')
                    elif fieldup[1][0:4] == "COLI": 
                        try:
                            hlist =fieldup[2:5]
                            vlist = []
                            for ilp in range(len(fieldup)): 
                                if fieldup[ilp][0:4] == "DIST":
                                    vlist.append(float(fieldup[ilp+1]))
                                if fieldup[ilp][0:4] == "SCAL":
                                    vlist.append(float(fieldup[ilp+1]))
                            vlist.append(float(0.0))
                            topology['lpics'].append([fieldup[1][0:4],hlist,vlist])
                        except (IndexError, ValueError):    
                            raise ValueError('Topology file LONEPAIR line has the wrong format. Ignored reading.')
                    elif fieldup[1][0:4] == "FIXE":
                        try:
                            topology['lpics'].append([[fieldup[1][0:4]],[],list(map(float,fieldup[2:5]))]) 
                        except (IndexError, ValueError):    
                            raise ValueError('Topology file LONEPAIR line has the wrong format. Ignored reading.')
                    elif fieldup[1][0:4] == "CENT":
                        try:
                            topology['lpics'].append([fieldup[1][0:4],fieldup[2:],[]])
                        except (IndexError, ValueError):    
                            raise ValueError('Topology file LONEPAIR line has the wrong format. Ignored reading.')
    
                # Anisotropy list dictionary
                elif ftype == "ANIS": 
                    try:
                        hlist = field[1:5]
                        for ilp in range(len(fieldup)): 
                            if fieldup[ilp][0:3] == "A11":
                                a11 =  float(fieldup[ilp+1])
                            if fieldup[ilp][0:3] == "A22":
                                a22 =  float(fieldup[ilp+1])
                        if a11 == 0.0 or a22 == 0.0:
                            raise ValueError('0.0 is invalid force constant. Correct the anisotropy line.')
                        vlist = [a11,a22] 
                        topology['anisotropies'].append([hlist,vlist])
                    except (IndexError, ValueError,KeyError):    
                        raise ValueError('Topology file Anisotropy line has the wrong format. Ignored reading.')
        #### Reading done ###
    
        if found:
            onlyatomsnlps = [at for at in topology['atomindex_byId'].keys() if at[0].upper() != "D"]
            onlyatoms     = [at for at in onlyatomsnlps if at[0:2].upper() != "LP" ]
            topology['ntopatoms']     = len(topology['atomindex_byId'].keys())
            topology['natoms']        = len(onlyatoms)
    
            #Check for residue total charge
            if abs( sum(topology['atomcharges']) - topology['rescharge'] ) > 0.00001:
                print('Warning: Your total charge does not match the sum of partial charges.')
    
        elif not found:
            print ('Could not find the residue name: '+topology['resname']+' in the provided topology file.')
        filein.close()
        return topology

    def loadCharmmTopology(self,resname,resitopfile,bio=True): 
        """
        Load topology information using CHARMM force field and store it as attributes.
        
        Args:
            resname (str): Name of the residue.
            resitopfile (str): Path to the CHARMM topology file.
        """

        topology = self.readCharmmTopology(resname,resitopfile,bio)
        for key,value in topology.items():
            self._assignattr(key,value)
    
    def _assignattr(self, key, value):
        """
        Private method to assign attributes to the class instance.

        Args:
            key: The attribute name.
            value: The value to assign to the attribute.
        """
        setattr(self, key, value)  

class WriteCharmmTopology():
    @staticmethod
    def rtf_header(header):
        text = ""
        if header:
            header=split_line(header)
            for h in header:
                text += "* "+ h + "\n"
        else:
            text="* Written by: amolkit\n"
        text += "*\n"
        return text
    
    @staticmethod
    def rtf_ioformat(extended=True):
        if extended: text="ioformat extended"
        return text
    
    @staticmethod
    def rtf_readinstruct(append=True):
        if append:
            text="read rtf card append"
        else:
            text="read rtf card"
        return text
    
    @staticmethod
    def rtf_version(version):
        text =  str(version)
        return text
    
    @staticmethod
    def rtf_autogenangdih(drude=True):
        if drude: 
            text = "AUTO ANGL DIHE DRUD"
        else:    
            text = "AUTO ANGL DIHE"
        return text    
    
    @staticmethod
    def rtf_comment(comment):
        text = ""          
        if comment:
            comment = split_line(comment)
            for h in comment:
                text += "! "+ h + "\n"
        else:
            text="! Written by: amolkit\n"
        return text
    
    @staticmethod
    def rtf_resname(resname,charge,charge_penalty=0.0,param_penalty=0.0):
        text=f"RESI {resname:<10} {charge:>5.3f} ! param penalty= {param_penalty:>6.3f} ; charge penalty= {charge_penalty:>4.1f}"
        return text
    
    @staticmethod
    def rtf_drudeatomline(topology,name):
        text = "ATOM {atomname:<6} {atomtype:<8} {charge:>10.3f} ALPHA {alpha:>6.4f} THOLE {thole:>6.4f} ! PENALTY {penalty:<6}".format(
                atomname=name,
                atomtype=topology['atomtype_byId'][name],
                charge=topology['atomcharge_byId'][name],
                alpha=topology['atomalpha_byId'][name],
                thole=topology['atomthole_byId'][name],
                penalty=topology['atompenalty_byId'][name])
        return text

    @staticmethod
    def rtf_additiveatomline(topology,name):
        text = "ATOM {atomname:<6} {atomtype:<8} {charge:>10.3f} ! PENALTY {penalty:<6}".format(
               atomname=name,
               atomtype=topology['atomtype_byId'][name],
               charge=topology['atomcharge_byId'][name],
               penalty=topology['atompenalty_byId'][name])
        return text

    @staticmethod
    def rtf_bond(bt1,bt2):
        text=f"BOND {bt1}  {bt2}"
        return text
        
    @staticmethod
    def rtf_improper(bt1,bt2,bt3,bt4): #topology):
        text=f"IMPR {bt1}  {bt2}  {bt3}  {bt4}"
        return text

    @staticmethod
    def rtf_allLonepairICs(topology):
        text = ""
        for value in topology["lpics"]:
            if not isinstance(value, (list, tuple)) or len(value) < 3:
                print(f"Warning: Invalid format for lone pair IC '{key}': {value}")
                continue  

            if value[0][:4] in ["RELA", "BISE"]:
                joined_values = " ".join(str(v) for v in value[1])
                text += f"LONE {value[0]} {joined_values} DIST {value[2][0]:>6.4f} ANGLE {value[2][1]:>7.2f} DIHE {value[2][2]:7.2f}\n"
            elif value[0][:4] == "COLI":
                joined_values = " ".join(str(v) for v in value[1])
                text += f"LONE {value[0]} {joined_values} DIST {value[2][0]:>6.4f} SCAL {value[2][1]:>7.1f}\n" 
        return text

    @staticmethod
    def rtf_end():
        text = "END"
        return text


def main():
    parser = argparse.ArgumentParser(description='Converts CHARMM Additve stream file into Drude stream file.')
    parser.add_argument('-r','--resi', type=str,help='Provide residue name/s for which drude str has to be created.',required=True)
    parser.add_argument('-t','--toppar', type=str,nargs='+',help='Provide file/s which contains the topology of provided residues.',required=True)
    parser.add_argument('-p','--parfile', type=str,help='Provide file/s which contains all drude parameters.',required=True)
    parser.add_argument('-walp','--wrtallprm', action='store_true',default=False,help='Insert this flag if you want to output all parameters')
    parser.add_argument('-o','--outname', type=str,help='Provide desired output name of drude str file.')
    args = parser.parse_args()
    #print (args.wrtallprm)
    cgenfftodrude(args.resi,args.toppar,args.parfile,args.wrtallprm,args.outname)

if __name__=="__main__":
   main()

# DO NOT DELETE
#import os,sys
#
#f1=sys.argv[1]
#f2=sys.argv[2]
#
#dictd={}
#with open(f1,"r") as fin:
#    fins = fin.readlines()
#for line in fins:
#    field = line.split()
#    if len(field) > 0 and field[0] == "MASS":
#        dictd.update({field[2]:[0.0,0.0,0,0]})
#    
#with open(f2,"r") as fin1:
#    fin1s = fin1.readlines()
#for line in fin1s:
#    field = line.split()
#    field = list(map(lambda x:x.strip(),field))
#    if len(field) > 0 and field[0] == "ATOM":
#        if "ALPHA" in field:
#            dictd[field[2]][0] = dictd[field[2]][0] + float(field[field.index("ALPHA")+1].strip("!"))
#            dictd[field[2]][2] = dictd[field[2]][2] + 1
#        if "THOLE" in field:
#            dictd[field[2]][1] = dictd[field[2]][1] + float(field[field.index("THOLE")+1].strip("!")) 
#            dictd[field[2]][3] = dictd[field[2]][3] + 1
#           
#for key,value in dictd.items():
#    if value[2] != 0:
#        dictd[key][0] = round(value[0]/value[2],4)
#    if value[3] != 0:
#        dictd[key][1] = round(value[1]/value[3],4)
#    def multiple_replace(self,dict,text):
#        regex = re.compile("(%s)" % "|".join(map(re.escape, list(dict.keys()))))
#        return regex.sub(lambda mo: dict[mo.string[mo.start():mo.end()]], text)

#
#NONBONDED nbxmod  5 atom vatom cdiel vdistance switch vswitch - 
#cutnb 16.0 ctofnb 12.0 ctonnb 10.0 eps 1.0 e14fac 1.0 wmin 1.5 



import sys,os
from setuptools import setup, find_packages, Command

class PyTest(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import subprocess
        errno = subprocess.call([sys.executable, '-m', 'pytest'])
        raise SystemExit(errno)

try:
     with open("README.md","r") as f:
          long_description=f.read()
except:
     long_description=""
     pass

setup(name='amolkit', 
      version='1.0.12', 
      description='Library for extracting molecule information', 
      long_description=long_description,
      url='https://github.com/anmolecule/amolkit', 
      author='Anmol Kumar', 
      author_email='anmolecule@gmail.com', 
      license='BSD 3-Clause "New" or "Revised" License',
      packages=find_packages(), 
      cmdclass={'test': PyTest},  
      zip_safe=False)

